##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest

from zope.annotation.tests.annotations import AnnotationsTestBase


class AttributeAnnotationsTest(AnnotationsTestBase, unittest.TestCase):

    def setUp(self):
        from zope.interface import implementer
        from zope.testing import cleanup

        from zope.annotation.attribute import AttributeAnnotations
        from zope.annotation.interfaces import IAttributeAnnotatable

        cleanup.setUp()

        @implementer(IAttributeAnnotatable)
        class Dummy:
            pass

        self.obj = Dummy()
        self.annotations = AttributeAnnotations(self.obj)

    def tearDown(self):
        from zope.testing import cleanup
        cleanup.tearDown()

    def testInterfaceVerifies(self):
        super().testInterfaceVerifies()
        self.assertIs(self.obj, self.annotations.__parent__)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Attribute Annotations implementation"""
import logging
from collections.abc import MutableMapping as DictMixin


try:
    from BTrees.OOBTree import OOBTree as _STORAGE
except ImportError:  # pragma: no cover
    logging.getLogger(__name__).warning(
        'BTrees not available: falling back to dict for attribute storage')
    _STORAGE = dict

from zope import component
from zope import interface
from zope.annotation import interfaces


_EMPTY_STORAGE = _STORAGE()


@interface.implementer(interfaces.IAnnotations)
@component.adapter(interfaces.IAttributeAnnotatable)
class AttributeAnnotations(DictMixin):
    """Store annotations on an object

    Store annotations in the `__annotations__` attribute on a
    `IAttributeAnnotatable` object.
    """

    # Yes, there's a lot of repetition of the `getattr` call,
    # but that turns out to be the most efficient for the ways
    # instances are typically used without sacrificing any semantics.
    # See https://github.com/zopefoundation/zope.annotation/issues/8
    # for a discussion of alternatives (which included functools.partial,
    # a closure, capturing the annotations in __init__, and versions
    # with getattr and exceptions).

    def __init__(self, obj, context=None):
        self.obj = obj

    @property
    def __parent__(self):
        return self.obj

    def __bool__(self):
        return bool(getattr(self.obj, '__annotations__', 0))

    def get(self, key, default=None):
        """See zope.annotation.interfaces.IAnnotations"""
        annotations = getattr(self.obj, '__annotations__', _EMPTY_STORAGE)
        return annotations.get(key, default)

    def __getitem__(self, key):
        annotations = getattr(self.obj, '__annotations__', _EMPTY_STORAGE)
        return annotations[key]

    def keys(self):
        annotations = getattr(self.obj, '__annotations__', _EMPTY_STORAGE)
        return annotations.keys()

    def __iter__(self):
        annotations = getattr(self.obj, '__annotations__', _EMPTY_STORAGE)
        return iter(annotations)

    def __len__(self):
        annotations = getattr(self.obj, '__annotations__', _EMPTY_STORAGE)
        return len(annotations)

    def __setitem__(self, key, value):
        """See zope.annotation.interfaces.IAnnotations"""
        try:
            annotations = self.obj.__annotations__
        except AttributeError:
            annotations = self.obj.__annotations__ = _STORAGE()

        annotations[key] = value

    def __delitem__(self, key):
        """See zope.app.interfaces.annotation.IAnnotations"""
        try:
            annotation = self.obj.__annotations__
        except AttributeError:
            raise KeyError(key)

        del annotation[key]



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""General Annotations Tests

All objects implementing 'IAnnotations' should pass these tests. They might be
used as base tests for real implementations.
"""
from zope.interface.verify import verifyObject

from zope.annotation.interfaces import IAnnotations


class AnnotationsTestBase:
    """Test the IAnnotations interface.

    The test case class expects the 'IAnnotations' implementer to be in
    'self.annotations'.
    """

    annotations = None
    obj = None

    def test_nonzero(self):
        self.assertFalse(self.annotations)
        self.annotations['unittest'] = self.obj
        self.assertTrue(self.annotations)
        del self.annotations['unittest']
        self.assertFalse(self.annotations)

    def testInterfaceVerifies(self):
        verifyObject(IAnnotations, self.annotations)

    def testStorage(self):
        # test __getitem__
        self.annotations['unittest'] = self.obj
        res = self.annotations['unittest']
        self.assertEqual(self.obj, res)

    def testGetitemException(self):
        # test __getitem__ raises exception on unknown key
        self.assertRaises(KeyError, self.annotations.__getitem__, 'randomkey')

    def testGet(self):
        # test get
        self.annotations['unittest'] = self.obj
        res = self.annotations.get('unittest')
        self.assertEqual(self.obj, res)

    def testGetNoSet(self):
        # test get with no set
        res = self.annotations.get('randomkey')
        self.assertEqual(None, res)

    def testGetDefault(self):
        # test get returns default
        res = self.annotations.get('randomkey', 'default')
        self.assertEqual('default', res)

    def testDel(self):
        self.annotations['unittest'] = self.obj
        del self.annotations['unittest']
        self.assertEqual(None, self.annotations.get('unittest'))
        self.assertEqual([], list(self.annotations.keys()))
        self.assertEqual([], list(self.annotations))
        self.assertEqual(0, len(self.annotations))

    def testDelRaisesKeyError(self):
        self.assertRaises(KeyError, self.annotations.__delitem__, 'unittest')

    def testKeys(self):
        self.assertEqual([], list(self.annotations.keys()))
        self.annotations['unittest'] = self.obj
        self.assertEqual(1, len(self.annotations))
        self.assertEqual(['unittest'], list(self.annotations.keys()))

    def testIter(self):
        self.assertEqual([], list(self.annotations))
        self.annotations['unittest'] = self.obj
        self.assertEqual(1, len(self.annotations))
        self.assertEqual(['unittest'], list(self.annotations))
        self.assertEqual(1, len(self.annotations))
